@@ -18,6 +18,14 @@
 
 var callbacks = {};
 
+// from https://stackoverflow.com/questions/105034/create-guid-uuid-in-javascript
+function generateGuid() {
+    return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(c) {
+        const r = Math.random()*16|0, v = c == 'x' ? r : (r&0x3|0x8);
+        return v.toString(16);
+    });
+}
+
 function addCallback(subsystem, action, callback)
 {
     if (!callbacks[subsystem]) {
@@ -26,15 +34,6 @@ function addCallback(subsystem, action, callback)
     callbacks[subsystem][action] = callback;
 }
 
-function sendMessage(subsystem, action, payload)
-{
-    (chrome.extension.sendMessage || browser.runtime.sendMessage)({
-        subsystem: subsystem,
-        action: action,
-        payload: payload
-    });
-}
-
 function executeScript(script) {
     var element = document.createElement('script');
     element.innerHTML = '('+ script +')();';
@@ -71,6 +70,14 @@ storage.get(DEFAULT_EXTENSION_SETTINGS, function (items) {
             loadMediaSessionsShim();
         }
     }
+
+    if (items.purpose.enabled) {
+        sendMessage("settings", "getSubsystemStatus").then((status) => {
+            if (status && status.purpose) {
+                loadPurpose();
+            }
+        });
+    }
 });
 
 // BREEZE SCROLL BARS
@@ -138,16 +145,9 @@ html::-webkit-scrollbar-corner {
 // ------------------------------------------------------------------------
 //
 
-// we give our transfer div a "random id" for privacy
-// from https://stackoverflow.com/questions/105034/create-guid-uuid-in-javascript
-var mediaSessionsTransferDivId ='xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(c) {
-    var r = Math.random()*16|0, v = c == 'x' ? r : (r&0x3|0x8);
-    return v.toString(16);
-});
-
 // also give the function a "random" name as we have to have it in global scope to be able
 // to invoke callbacks from outside, UUID might start with a number, so prepend something
-var mediaSessionsClassName = "f" + mediaSessionsTransferDivId.replace(/-/g, "");
+const mediaSessionsClassName = "f" + generateGuid().replace(/-/g, "");
 
 var activePlayer;
 // When a player has no duration yet, we'll wait for it becoming known
@@ -163,6 +163,9 @@ var players = [];
 
 var pendingSeekingUpdate = 0;
 
+var titleTagObserver = null;
+var oldPageTitle = "";
+
 addCallback("mpris", "play", function () {
     playerPlay();
 });
@@ -241,6 +244,16 @@ addCallback("mpris", "previous", function () {
     }
 });
 
+addCallback("mpris", "setFullscreen", (message) => {
+    if (activePlayer) {
+        if (message.fullscreen) {
+            activePlayer.requestFullscreen();
+        } else {
+            document.exitFullscreen();
+        }
+    }
+});
+
 addCallback("mpris", "setPosition", function (message) {
     if (activePlayer) {
         activePlayer.currentTime = message.position;
@@ -301,7 +314,7 @@ function setPlayerActive(player) {
     // Ignore short sounds, they are most likely a chat notification sound
     // A stream has a duration of Infinity
     // Note that "NaN" is also not finite but we already returned earlier for that
-    if (isFinite(player.duration) && player.duration > 0 && player.duration < 5) {
+    if (isFinite(player.duration) && player.duration > 0 && player.duration < 8) {
         return;
     }
 
@@ -312,6 +325,7 @@ function setPlayerActive(player) {
     // and then we would have ignored the metadata signal because there was no player
     sendMessage("mpris", "playing", {
         mediaSrc: player.src,
+        pageTitle: document.title,
         duration: player.duration,
         currentTime: player.currentTime,
         playbackRate: player.playbackRate,
@@ -319,8 +333,37 @@ function setPlayerActive(player) {
         muted: player.muted,
         loop: player.loop,
         metadata: playerMetadata,
-        callbacks: playerCallbacks
+        callbacks: playerCallbacks,
+        fullscreen: document.fullscreenElement !== null,
+        canSetFullscreen: player.tagName.toLowerCase() === "video"
     });
+
+    if (!titleTagObserver) {
+
+        // Observe changes to the <title> tag in case it is updated after the player has started playing
+        let titleTag = document.querySelector("head > title");
+        if (titleTag) {
+            oldPageTitle = titleTag.innerText;
+
+            titleTagObserver = new MutationObserver((mutations) => {
+                mutations.forEach((mutation) => {
+                    const pageTitle = mutation.target.textContent;
+                    if (pageTitle && oldPageTitle !== pageTitle) {
+                        sendMessage("mpris", "titlechange", {
+                            pageTitle: pageTitle
+                        });
+                    }
+                    oldPageTitle = pageTitle;
+                });
+            });
+
+            titleTagObserver.observe(titleTag, {
+                childList: true, // text content is technically a child node
+                subtree: true,
+                characterData: true
+            });
+        }
+    }
 }
 
 function sendPlayerGone() {
@@ -334,6 +377,11 @@ function sendPlayerGone() {
     playerMetadata = {};
     playerCallbacks = [];
     sendMessage("mpris", "gone");
+
+    if (titleTagObserver) {
+        titleTagObserver.disconnect();
+        titleTagObserver = null;
+    }
 }
 
 function sendPlayerInfo(player, event, payload) {
@@ -532,32 +580,11 @@ function loadMpris() {
         });
     });
 
-    observer.observe(document.documentElement, {
+    observer.observe(document, {
         childList: true,
         subtree: true
     });
 
-    // Observe changes to the <title> tag in case it is updated after the player has started playing
-    var titleTag = document.querySelector("head > title");
-    if (titleTag) {
-        var titleObserver = new MutationObserver(function (mutations) {
-            mutations.forEach(function (mutation) {
-                var pageTitle = mutation.target.textContent;
-                if (pageTitle) {
-                    sendMessage("mpris", "titlechange", {
-                        pageTitle: pageTitle
-                    });
-                }
-            });
-        });
-
-        titleObserver.observe(titleTag, {
-            childList: true, // text content is technically a child node
-            subtree: true,
-            characterData: true
-        });
-    }
-
     window.addEventListener("beforeunload", function () {
         // about to navigate to a different page, tell our extension that the player will be gone shortly
         // we listen for tab closed in the extension but we don't for navigating away as URL change doesn't
@@ -572,14 +599,19 @@ function loadMpris() {
     document.addEventListener("DOMContentLoaded", function() {
         registerAllPlayers();
     });
+
+    document.addEventListener("fullscreenchange", () => {
+        if (activePlayer) {
+            sendPlayerInfo(activePlayer, "fullscreenchange", {
+                fullscreen: document.fullscreenElement !== null
+            });
+        }
+    });
 }
 
 // This adds a shim for the Chrome media sessions API which is currently only supported on Android
 // Documentation: https://developers.google.com/web/updates/2017/02/media-session
 // Try it here: https://googlechrome.github.io/samples/media-session/video.html
-//
-// TODO Forward mpris calls to the actionHandlers on the page
-// previoustrack, nexttrack, seekbackward, seekforward, play, pause
 
 // Bug 379087: Only inject this stuff if we're a proper HTML page
 // otherwise we might end up messing up XML stuff
@@ -589,22 +621,78 @@ function loadMpris() {
 // tagName always returned "HTML" for me but I wouldn't trust it always being uppercase
 function loadMediaSessionsShim() {
     if (document.documentElement.tagName.toLowerCase() === "html") {
+
+        window.addEventListener("pbiMprisMessage", (e) => {
+            let data = e.detail || {};
+
+            let action = data.action;
+            let payload = data.payload;
+
+            switch (action) {
+            case "metadata":
+                playerMetadata = {};
+
+                if (typeof payload !== "object") {
+                    return;
+                }
+
+                playerMetadata = payload;
+                sendMessage("mpris", "metadata", payload);
+
+                return;
+
+            case "playbackState":
+                if (!["none", "paused", "playing"].includes(payload)) {
+                    return;
+                }
+
+                playerPlaybackState = payload;
+
+                if (!activePlayer) {
+                    return;
+                }
+
+                if (playerPlaybackState === "playing") {
+                    playerPlaying(activePlayer);
+                } else if (playerPlaybackState === "paused") {
+                    playerPaused(activePlayer);
+                }
+
+                return;
+
+            case "callbacks":
+                if (Array.isArray(payload)) {
+                    playerCallbacks = payload;
+                } else {
+                    playerCallbacks = [];
+                }
+                sendMessage("mpris", "callbacks", playerCallbacks);
+
+                return;
+            }
+        });
+
         executeScript(`
             function() {
                 ${mediaSessionsClassName} = function() {};
-                ${mediaSessionsClassName}.transferItem = null;
                 ${mediaSessionsClassName}.callbacks = {};
                 ${mediaSessionsClassName}.metadata = null;
                 ${mediaSessionsClassName}.playbackState = "none";
                 ${mediaSessionsClassName}.sendMessage = function(action, payload) {
-                    this.transferItem.innerText = JSON.stringify({action: action, payload: payload});
-
-                    var event = document.createEvent('CustomEvent');
-                    event.initEvent('payloadChanged', true, true);
-                    this.transferItem.dispatchEvent(event);
+                    let event = new CustomEvent("pbiMprisMessage", {
+                        detail: {
+                            action: action,
+                            payload: payload
+                        }
+                    });
+                    window.dispatchEvent(event);
                 };
                 ${mediaSessionsClassName}.executeCallback = function (action) {
-                    this.callbacks[action]();
+                    let details = {
+                        action: action
+                        // for seekforward, seekbackward, seekto there's additional information one would need to add
+                    };
+                    this.callbacks[action](details);
                 };
 
                 if (!navigator.mediaSession) {
@@ -615,13 +703,6 @@ function loadMediaSessionsShim() {
 
                 var oldSetActionHandler = navigator.mediaSession.setActionHandler || noop;
                 navigator.mediaSession.setActionHandler = function (name, cb) {
-                    // Call the original native implementation
-                    // "call()" is needed as the real setActionHandler is a class member
-                    // and calling it directly is illegal as it lacks the context
-                    // We'll register the callback for ourself after this since it may
-                    // throw for unsupported callback names.
-                    var ret = oldSetActionHandler.call(navigator.mediaSession, name, cb);
-
                     if (cb) {
                         ${mediaSessionsClassName}.callbacks[name] = cb;
                     } else {
@@ -629,7 +710,12 @@ function loadMediaSessionsShim() {
                     }
                     ${mediaSessionsClassName}.sendMessage("callbacks", Object.keys(${mediaSessionsClassName}.callbacks));
 
-                    return ret;
+                    // Call the original native implementation
+                    // "call()" is needed as the real setActionHandler is a class member
+                    // and calling it directly is illegal as it lacks the context
+                    // This may throw for unsupported actions but we registered the callback
+                    // ourselves before
+                    return oldSetActionHandler.call(navigator.mediaSession, name, cb);
                 };
 
                 Object.defineProperty(navigator.mediaSession, "metadata", {
@@ -682,8 +768,8 @@ function loadMediaSessionsShim() {
         // mechanism. Let's see how this goes :D
 
         executeScript(`function() {
-                var oldCreateElement = document.createElement;
-                document.createElement = function () {
+                var oldCreateElement = Document.prototype.createElement;
+                Document.prototype.createElement = function() {
                     var createdTag = oldCreateElement.apply(this, arguments);
 
                     var tagName = arguments[0];
@@ -701,65 +787,179 @@ function loadMediaSessionsShim() {
         `);
 
         // We also briefly add items created as new Audio() to the DOM so we can control it
-        // similar to the document.createElement hack above
-        executeScript(`function() {
+        // similar to the document.createElement hack above since we cannot share variables
+        // between the actual website and the background script despite them sharing the same DOM
+        // HACK When removing a media object from DOM it is paused, so what we do here is once the
+        // player loaded some data we add it (doesn't work earlier since it cannot pause when
+        // there's nothing loaded to pause) to the DOM and before we remove it, we note down that
+        // we will now get a paused event because of that. When we get it, we just play() the player
+        // so it continues playing :-)
+        if (IS_FIREFOX) {
+            // Firefox enforces Content-Security-Policy also for scripts injected by the content-script
+            // This causes our executeScript calls to fail for pages like Nextcloud
+            // It also doesn't seem to have the aggressive autoplay prevention Chrome has,
+            // so the horrible replyAfterRemoval hack from above isn't copied into this
+            // See Bug 411148: Music playing from the ownCloud Music app does not show up
+            var oldAudio = window.Audio;
+            exportFunction(function() {
+                var createdAudio = new (Function.prototype.bind.apply(oldAudio, arguments));
+
+                (document.head || document.documentElement).appendChild(createdAudio);
+                createdAudio.parentNode.removeChild(createdAudio);
+
+                return createdAudio;
+            }, window, {defineAs: "Audio"});
+        } else {
+            executeScript(`function() {
                 var oldAudio = window.Audio;
                 window.Audio = function () {
                     var createdAudio = new (Function.prototype.bind.apply(oldAudio, arguments));
 
-                    (document.head || document.documentElement).appendChild(createdAudio);
-                    createdAudio.parentNode.removeChild(createdAudio);
+                    createdAudio.registerInDom = function() {
+                        (document.head || document.documentElement).appendChild(createdAudio);
+                        createdAudio.pausedBecauseOfDomRemoval = true;
+                        createdAudio.parentNode.removeChild(createdAudio);
+
+                        createdAudio.removeEventListener("loadeddata", createdAudio.registerInDom);
+                    };
+
+                    createdAudio.replayAfterRemoval = function() {
+                        if (createdAudio.pausedBecauseOfDomRemoval) {
+                            if (createdAudio.paused) {
+                                createdAudio.play();
+                            }
+                            delete createdAudio.pausedBecauseOfDomRemoval;
+
+                            createdAudio.removeEventListener("pause", createdAudio.replayAfterRemoval);
+                        }
+                    };
+
+                    createdAudio.addEventListener("loadeddata", createdAudio.registerInDom);
+                    createdAudio.addEventListener("pause", createdAudio.replayAfterRemoval);
 
                     return createdAudio;
                 };
-            }
-        `);
-
-        // now the fun part of getting the stuff from our page back into our extension...
-        // cannot access extensions from innocent page JS for security
-        var transferItem = document.createElement("div");
-        transferItem.setAttribute("id", mediaSessionsTransferDivId);
-
-        // Add the element to the DOM so the page sees it
-        (document.head || document.documentElement).appendChild(transferItem);
-
-        // now keep a reference to it inside the page
-        executeScript(`
-            function() {
-                ${mediaSessionsClassName}.transferItem = document.getElementById('${mediaSessionsTransferDivId}');
-            }
-        `);
-
-        // and remove it again so the website isn't confused
-        transferItem.parentNode.removeChild(transferItem);
-
-        transferItem.addEventListener('payloadChanged', function() {
-            var json = JSON.parse(this.innerText);
-
-            var action = json.action
-
-            if (action === "metadata") {
-                // FIXME filter metadata, this stuff comes from a hostile environment after all
-
-                playerMetadata = json.payload;
-
-                sendMessage("mpris", "metadata", json.payload);
-            } else if (action === "playbackState") {
-                var playbackState = json.payload;
-
-                playerPlaybackState = playbackState;
-                if (activePlayer) {
-                    if (playbackState === "playing") {
-                        playerPlaying(activePlayer);
-                    } else if (playbackState === "paused") {
-                        playerPaused(activePlayer);
-                    }
-                }
-
-            } else if (action === "callbacks") {
-                playerCallbacks = json.payload;
-                sendMessage("mpris", "callbacks", json.payload);
-            }
-        });
+            }`);
+        }
     }
 }
+
+// PURPOSE / WEB SHARE API
+// ------------------------------------------------------------------------
+//
+const purposeTransferClassName = "p" + generateGuid().replace(/-/g, "");
+
+var purposeLoaded = false;
+function loadPurpose() {
+    if (purposeLoaded) {
+        return;
+    }
+
+    purposeLoaded = true;
+
+    // navigator.share must only be defined in secure (https) context
+    if (!window.isSecureContext) {
+        return;
+    }
+
+     window.addEventListener("pbiPurposeMessage", (e) => {
+        const data = e.detail || {};
+
+        const action = data.action;
+        const payload = data.payload;
+
+        if (action !== "share") {
+            return;
+        }
+
+        sendMessage("purpose", "share", payload).then((response) => {
+            executeScript(`
+                function() {
+                    ${purposeTransferClassName}.pendingResolve();
+                }
+            `);
+        }, (err) => {
+            // Deliberately not giving any more details about why it got rejected
+            executeScript(`
+                function() {
+                    ${purposeTransferClassName}.pendingReject(new DOMException("Share request aborted", "AbortError"));
+                }
+            `);
+        }).finally(() => {
+            executeScript(`
+                function() {
+                    ${purposeTransferClassName}.reset();
+                }
+            `);
+        });;
+    });
+
+    executeScript(`
+        function() {
+            ${purposeTransferClassName} = function() {};
+            let transfer = ${purposeTransferClassName};
+            transfer.reset = () => {
+                transfer.pendingResolve = null;
+                transfer.pendingReject = null;
+            };
+            transfer.reset();
+
+            if (!navigator.canShare) {
+                navigator.canShare = (data) => {
+                    if (!data) {
+                        return false;
+                    }
+
+                    if (data.title === undefined && data.text === undefined && data.url === undefined) {
+                        return false;
+                    }
+
+                    if (data.url) {
+                        // check if URL is valid
+                        try {
+                            new URL(data.url, document.location.href);
+                        } catch (e) {
+                            return false;
+                        }
+                    }
+
+                    return true;
+                }
+            }
+
+            if (!navigator.share) {
+                navigator.share = (data) => {
+                    return new Promise((resolve, reject) => {
+                        if (!navigator.canShare(data)) {
+                            return reject(new TypeError());
+                        }
+
+                        if (data.url) {
+                            // validity already checked in canShare, hence no catch
+                            data.url = new URL(data.url, document.location.href).toString();
+                        }
+
+                        if (!window.event || !window.event.isTrusted) {
+                            return reject(new DOMException("navigator.share can only be called in response to user interaction", "NotAllowedError"));
+                        }
+
+                        if (transfer.pendingResolve || transfer.pendingReject) {
+                            return reject(new DOMException("A share is already in progress", "AbortError"));
+                        }
+
+                        transfer.pendingResolve = resolve;
+                        transfer.pendingReject = reject;
+
+                        const event = new CustomEvent("pbiPurposeMessage", {
+                            detail: {
+                                action: "share",
+                                payload: data
+                            }
+                        });
+                        window.dispatchEvent(event);
+                    });
+                };
+            }
+        }
+    `);
+}
